︠78d5f1ec-52b4-4050-a9f5-c53ca84b16d8i︠
%html
<p>I stayed up a bit and wrote an LU decomposition function for Sage.&nbsp; You can see the actual code that is going into Sage soon at <a href="http://trac.sagemath.org/sage_trac/ticket/3048">http://trac.sagemath.org/sage_trac/ticket/3048</a> or <a href="http://trac.sagemath.org/sage_trac/attachment/ticket/3048/lu_decomposition.patch">http://trac.sagemath.org/sage_trac/attachment/ticket/3048/lu_decomposition.patch</a>.&nbsp; In particular, you might notice that the new LU decomposition code is used in calculating the determinant of a matrix, solving a system of inequalities, and finding the inverse of a matrix.</p>

︡153736dd-15a5-4bd0-aa7f-953dded7cda6︡{"html": "<p>I stayed up a bit and wrote an LU decomposition function for Sage.&nbsp; You can see the actual code that is going into Sage soon at <a href=\"http://trac.sagemath.org/sage_trac/ticket/3048\">http://trac.sagemath.org/sage_trac/ticket/3048</a> or <a href=\"http://trac.sagemath.org/sage_trac/attachment/ticket/3048/lu_decomposition.patch\">http://trac.sagemath.org/sage_trac/attachment/ticket/3048/lu_decomposition.patch</a>.&nbsp; In particular, you might notice that the new LU decomposition code is used in calculating the determinant of a matrix, solving a system of inequalities, and finding the inverse of a matrix.</p>"}︡
︠ad97dddd-4e70-4193-8a75-793817058fe7a︠
%auto
%cython
from sage.matrix.matrix cimport Matrix
def _plu_decomposition_classical(Matrix self, pivoting=None):
    """
    Return the PLU decomposition P, L, and U such that self=P*L*U.
    If self is an mxn matrix, then P is an mxm permutation matrix,
    L is a mxm unit lower-triangular matrix, and U is an mxn
    upper-triangular matrix.

    The decomposition is done over the fraction field of self.base_ring().

    EXAMPLES::

    """
    #tm = verbose('generic in-place LU decomposition on %s x %s matrix'%(self._nrows, self._ncols))
    cdef Py_ssize_t start_row, c, r, nr, nc, i, pivot_row
    # Default to partial pivoting for inexact fields
    R = self.base_ring().fraction_field()

    x = self.fetch('PLU')
    if x is not None and x[3]==pivoting:
        return x[0].matrix(), x[1], x[2]

    cdef Matrix A, L

    nr = self._nrows
    nc = self._ncols
    R = self.base_ring().fraction_field()
    A = self.change_ring(R).copy()
    L = A.new_matrix(nr, nr)

    from sage.groups.all import SymmetricGroup 
    S = SymmetricGroup(nr)
    p = S(1)

    start_row = 0
    pivots = []
    for c from 0 <= c < nc:
        #print "column: ", c
        if PyErr_CheckSignals(): raise KeyboardInterrupt
        pivot = 0
        pivot_row = start_row
        if pivoting=="partial":
            # Pick largest possible pivot in column c
            for r from start_row <= r < nr:
                if pivot < abs(A.get_unsafe(r,c)):
                    pivot_row = r
                    pivot = abs(A.get_unsafe(pivot_row, c))
        else:
            # Pick the first nonzero pivot in column c
            for r from start_row <= r < nr:
                pivot = A.get_unsafe(r,c)
                if pivot != 0:
                    pivot_row = r
                    break

        if pivot != 0:
            pivots.append(c)
            if pivot_row != start_row:
                A.swap_rows(pivot_row, start_row)
                L.swap_rows(pivot_row, start_row)
                p *= S((pivot_row+1, start_row+1))
            a_inverse = ~A.get_unsafe(start_row,c)
            for i from start_row < i < nr:
                if A.get_unsafe(i,c):
                    b = A.get_unsafe(i, c)*a_inverse
                    A.add_multiple_of_row(i, start_row, -b, c)
                    L.set_unsafe(i,start_row,b)
            start_row += 1

    L += L.parent().identity_matrix()
    self.cache('pivots', pivots)
    self.cache('PLU', (p,L,A,pivoting))
    #verbose('done with LU decomposition', tm)            
    return p.matrix(), L, A

def lu_decomposition(Matrix self, *args, **kwds):
    """
    Wrapper function for PLU decomposition.
    """
    return _plu_decomposition_classical(self,*args,**kwds)


def solver(Matrix self, Matrix B, check_rank=True):
    r"""
    If self is a matrix `A` of full rank, then this function
    returns a matrix `X` such that `A X = B`.
    
    .. seealso::

       :meth:`solve_right` and :meth:`solve_left`
    
    INPUT:
    
    
    -  ``B`` - a matrix
    
    -  ``check_rank`` - bool (default: True) (currently ignored)
    
    
    OUTPUT: matrix
    
    EXAMPLES::
    
        sage: A = matrix(QQ,3,[1,2,4,5,3,1,1,2,-1])
        sage: B = matrix(QQ,3,2,[1,5,1,2,1,5])
        sage: A._solve_right_nonsingular_square(B)
        [ -1/7 -11/7]
        [  4/7  23/7]
        [    0     0]
        sage: A._solve_right_nonsingular_square(B, check_rank=False)
        [ -1/7 -11/7]
        [  4/7  23/7]
        [    0     0]
        sage: X = A._solve_right_nonsingular_square(B, check_rank=False)
        sage: A*X == B
        True
    """
    cdef Matrix L,U
    cdef int i,j
    # Should use LU Decomposition!
    lu_decomposition(self)
    P,L,U,strategy=self.fetch('PLU')
    # Swap matrix rows of B using P
    B=B[(~P)(range(B._nrows))].copy()
    # Forward substitution
    for i from 1<=i<B._nrows:
        B[i] = B[i]-sum([L[i,j]*B[j] for j in range(i)])
    # Backwards substitution
    for i from B._nrows>i>=0:
        B[i] = (B[i]-sum([U[i,j]*B[j] for j in range(U._ncols-1,i,-1)]))/U[i,i]
    return B
    
    
def det(Matrix self):
    cdef Matrix u
    lu_decomposition(self) 
    p,l,u,strategy=self.fetch('PLU') 
    d = p.sign() 
    for i from 0<= i < self._nrows: 
        d *= u.get_unsafe(i,i) 
    self.cache('det', d) 
    return d
︡ecf6e8f7-c9d0-48c9-84cd-302a029fdf78︡︡
︠28b17ee9-a9a5-47b1-bcbe-ff51b2f714e6i︠
%html
<p>Let's try the example from the book.</p>

︡e9e1ac89-76db-4002-bd95-3f176587bc7e︡{"html": "<p>Let's try the example from the book.</p>"}︡
︠18785540-c13d-4485-9a16-4a88136be00a︠
A=matrix(4,5,[2,4,-1,5,-2, -4,-5,3,-8,1, 2, -5, -4, 1, 8, -6, 0, 7, -3, 1])
show(A)
︡bd381ad4-f384-471b-967b-4f28885e7be4︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrrr}\n2 & 4 & -1 & 5 & -2 \\\\\n-4 & -5 & 3 & -8 & 1 \\\\\n2 & -5 & -4 & 1 & 8 \\\\\n-6 & 0 & 7 & -3 & 1\n\\end{array}\\right)</div>"}︡
︠7564c3c7-48ab-44f3-9d52-a73f13eb764ci︠
%html
<p>If the following command doesn't work, evaluate the first cell.&nbsp; You can also look at the source code in that cell for the LU decomposition.&nbsp; It's slightly more general than the algorithm we talked about in class, but the underlying principles are exactly the same.&nbsp; The generalization came in when doing the pivoting and row swaps.</p>

︡23fd2557-bf49-4b29-b018-f40fa9bddc55︡{"html": "<p>If the following command doesn't work, evaluate the first cell.&nbsp; You can also look at the source code in that cell for the LU decomposition.&nbsp; It's slightly more general than the algorithm we talked about in class, but the underlying principles are exactly the same.&nbsp; The generalization came in when doing the pivoting and row swaps.</p>"}︡
︠611eb719-fa71-4801-8157-0696149dd77b︠
P,L,U=lu_decomposition(A)
show(P)
show(L)
show(U)
︡a4e51630-f188-46b3-9075-5d2807958eea︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrr}\n1 & 0 & 0 & 0 \\\\\n0 & 1 & 0 & 0 \\\\\n0 & 0 & 1 & 0 \\\\\n0 & 0 & 0 & 1\n\\end{array}\\right)</div>"}︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrr}\n1 & 0 & 0 & 0 \\\\\n-2 & 1 & 0 & 0 \\\\\n1 & -3 & 1 & 0 \\\\\n-3 & 4 & 2 & 1\n\\end{array}\\right)</div>"}︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrrr}\n2 & 4 & -1 & 5 & -2 \\\\\n0 & 3 & 1 & 2 & -3 \\\\\n0 & 0 & 0 & 2 & 1 \\\\\n0 & 0 & 0 & 0 & 5\n\\end{array}\\right)</div>"}︡
︠fc811413-7047-44b2-b9d9-af45c776362ai︠
%html
<p>Let's check that it is right.</p>

︡4224069f-f021-45eb-b474-5f551ea153e6︡{"html": "<p>Let's check that it is right.</p>"}︡
︠5472e3d3-4b00-4b6c-8c37-3b94f2d4aacd︠
P*L*U==A
︡58b429c1-f014-4beb-ae82-fea692655082︡{"stdout": "True"}︡
︠b81eb364-fe64-45be-b2eb-c4fc8e9626aai︠
%html
<p>Remember that Matlab gave us a different answer?&nbsp; That was because Matlab used partial pivoting (it chose the largest pivot and did a row swap).&nbsp; Partial pivoting helps minimize numerical errors.&nbsp; In Sage, because there is exact arithmetic (and we are dealing with exact numbers), we don't need to use partial pivoting.&nbsp; However, we can still ask for it anyway.</p>

︡877cef07-ece5-4954-839b-85795f970f1b︡{"html": "<p>Remember that Matlab gave us a different answer?&nbsp; That was because Matlab used partial pivoting (it chose the largest pivot and did a row swap).&nbsp; Partial pivoting helps minimize numerical errors.&nbsp; In Sage, because there is exact arithmetic (and we are dealing with exact numbers), we don't need to use partial pivoting.&nbsp; However, we can still ask for it anyway.</p>"}︡
︠2366e694-ecce-4db6-afc4-16e05161ab4f︠
P,L,U=lu_decomposition(A, pivoting="partial")
show(P)
show(L)
show(U)
︡4368a82b-5c50-4ae3-9947-8298cd234d4e︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrr}\n0 & 0 & 0 & 1 \\\\\n0 & 1 & 0 & 0 \\\\\n0 & 0 & 1 & 0 \\\\\n1 & 0 & 0 & 0\n\\end{array}\\right)</div>"}︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrr}\n1 & 0 & 0 & 0 \\\\\n\\frac{2}{3} & 1 & 0 & 0 \\\\\n-\\frac{1}{3} & 1 & 1 & 0 \\\\\n-\\frac{1}{3} & -\\frac{4}{5} & -\\frac{2}{15} & 1\n\\end{array}\\right)</div>"}︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrrr}\n-6 & 0 & 7 & -3 & 1 \\\\\n0 & -5 & -\\frac{5}{3} & -6 & \\frac{1}{3} \\\\\n0 & 0 & 0 & 6 & 8 \\\\\n0 & 0 & 0 & 0 & -\\frac{1}{3}\n\\end{array}\\right)</div>"}︡
︠b9135340-b7c9-4437-ab84-23980fca6853i︠
%html
<p>Notice that from $P$, we see that the first and last rows are swapped in order to get better pivots numerically.&nbsp; Let's still verify the results.</p>

︡51243adb-397a-4f2f-a1ca-129584ad0d68︡{"html": "<p>Notice that from $P$, we see that the first and last rows are swapped in order to get better pivots numerically.&nbsp; Let's still verify the results.</p>"}︡
︠76d1f38d-7dc9-45cc-b6f2-c305338f7545︠
P*L*U==A
︡65601e55-9d99-4a22-804e-dd267bace17e︡{"stdout": "True"}︡
︠53e6ce8a-62e3-4f7a-be3c-b01d3ee0ff59i︠
%html
<p>You might notice in other examples that the Matlab $P$ is still a bit different than the Sage $P$.&nbsp; That's because the Matlab $L$, $U$, and $P$ satisfy $LU=PA$, while the Sage results satisfy $PLU=A$.&nbsp; As we'll find out later in the course, basically the Matlab $P$ is the transpose of the Sage $P$.&nbsp; There are different conventions for where $P$ should go.&nbsp; Some other software does a $LUP=A$ decomposition.&nbsp; The Lapack software does a $PLU=A$ decomposition, like Sage.</p>
<p>$CC$ is the set of complex numbers, so here, we construct a random $4\times 5$ complex matrix.</p>

︡0fe2c21f-0b1e-41c5-a3d2-0e3ca6004d40︡{"html": "<p>You might notice in other examples that the Matlab $P$ is still a bit different than the Sage $P$.&nbsp; That's because the Matlab $L$, $U$, and $P$ satisfy $LU=PA$, while the Sage results satisfy $PLU=A$.&nbsp; As we'll find out later in the course, basically the Matlab $P$ is the transpose of the Sage $P$.&nbsp; There are different conventions for where $P$ should go.&nbsp; Some other software does a $LUP=A$ decomposition.&nbsp; The Lapack software does a $PLU=A$ decomposition, like Sage.</p>\n<p>$CC$ is the set of complex numbers, so here, we construct a random $4\\times 5$ complex matrix.</p>"}︡
︠2d4d4e07-f797-4fd8-b941-527092fadb96︠
a=random_matrix(CC,4,5)
show(a)
︡7ffba3d5-a598-4ed8-8c2f-c1bcb4e75c18︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrrr}\n0.229692748016513 - 0.845661923562603i & 0.287536701715982 + 0.236946603380285i & -0.185534572565047 - 0.289195161295678i & 0.381784037827768 - 0.869211099061085i & -0.0586247227789061 - 0.293832522677727i \\\\\n0.572588741325297 - 0.188787845429612i & -0.179735519225174 + 0.226375704374467i & 0.807922936521494 + 0.513268974988697i & -0.652204136019794 + 0.672635212571936i & -0.0441922790787861 - 0.0549952757621675i \\\\\n-0.630452932491459 + 0.132255191298339i & -0.154287888785182 + 0.924044425657864i & -0.533770069467662 + 0.866141745963207i & -0.341068398547977 - 0.107524013242364i & 0.722668222090292 - 0.299564064600013i \\\\\n-0.156088117010016 + 0.654210787293331i & 0.194396079337710 + 0.910673000800208i & 0.652747709221585 + 0.354757282827718i & -0.415673580873110 + 0.342933310778510i & -0.0176146945731379 - 0.323775635584187i\n\\end{array}\\right)</div>"}︡
︠f14c4f32-4518-48a8-a732-6882530c1415i︠
%html
<p>If we have a matrix that is not exact (like this floating point complex matrix), the function above knows automatically to do partial pivoting to minimize numerical error.</p>

︡f1e4ea49-270b-4085-a38a-d8c3fa6679b9︡{"html": "<p>If we have a matrix that is not exact (like this floating point complex matrix), the function above knows automatically to do partial pivoting to minimize numerical error.</p>"}︡
︠aa343c67-aef0-4887-9154-eb8f4657ee0d︠
P,L,U=lu_decomposition(a)
show(P)
show(L)
show(U)
︡1f82b24d-5ce1-4daa-9441-da86d3317ee9︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrr}\n1 & 0 & 0 & 0 \\\\\n0 & 0 & 1 & 0 \\\\\n0 & 1 & 0 & 0 \\\\\n0 & 0 & 0 & 1\n\\end{array}\\right)</div>"}︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrr}\n1.00000000000000 & 0 & 0 & 0 \\\\\n-0.334226714856104 - 0.654733841584202i & 1.00000000000000 & 0 & 0 \\\\\n0.379175796722956 + 0.574100355079673i & -0.000975088882645078 + 0.128358851883204i & 1.00000000000000 & 0 \\\\\n-0.767145822259989 + 0.0237916766424910i & 0.821587531362721 - 0.500109006499683i & -0.0481482063416919 - 0.705230791651185i & 1.00000000000000\n\\end{array}\\right)</div>"}︡{"html": "<div class=\"math\">\\left(\\begin{array}{rrrrr}\n0.229692748016513 - 0.845661923562603i & 0.287536701715982 + 0.236946603380285i & -0.185534572565047 - 0.289195161295678i & 0.381784037827768 - 0.869211099061085i & -0.0586247227789061 - 0.293832522677727i \\\\\n0 & -0.213322401451592 + 1.19149831981292i & -0.406434821225628 + 0.648009233808876i & 0.355635948235587 - 0.148070653655511i & 0.895456369941713 - 0.436154333329313i \\\\\n-5.55111512312578 \\times 10^{-17}i & -2.77555756156289 \\times 10^{-17} - 3.46944695195361 \\times 10^{-18}i & 0.795027522292052 + 0.782241618288173i & -1.31464120579083 + 0.737243267846046i & -0.245763678633482 - 0.0252896615076734i \\\\\n5.55111512312578 \\times 10^{-17} & 5.55111512312578 \\times 10^{-17} + 2.22044604925031 \\times 10^{-16}i & 0 & -0.944828403398494 - 0.925080408664571i & -0.581148284506794 + 0.0838337447132584i\n\\end{array}\\right)</div>"}︡
︠1472516d-f558-4d5e-8021-1cbefafeffa3i︠
%html
<p>Check the results.&nbsp; We can see that the difference between the two matrices is really small.</p>

︡cac1b894-cbb8-49d2-9c85-9963209e5009︡{"html": "<p>Check the results.&nbsp; We can see that the difference between the two matrices is really small.</p>"}︡
︠b91b459b-4a94-4195-9f6b-4f9eb74e9a7a︠
P*L*U-a
︡80a84466-6b0e-429f-91e5-078acffde830︡{"stdout": "[                      0                       0                       0                       0                       0]\n[                      0   -2.77555756156289e-17                       0                       0                       0]\n[                      0                       0                       0                       0                       0]\n[  -2.77555756156289e-17                       0                       0 -1.11022302462516e-16*I                       0]"}︡
︠c80c5bbf-fb7a-4f86-919c-93f0b0d31313︠
def make_symbolic_matrix(size):
    return matrix(size,[var('x%s%s'%(i,j)) for i in [0..size-1] for j in [0..size-1]])
︡05d7ffdc-6505-4916-92dd-c1e8db16a248︡︡
︠b2ccf4e8-16df-4d1b-95da-64aa7b7cf117︠
A=make_symbolic_matrix(5)
A
︡46f69b6d-67da-49d3-b505-cc4a2d9c0b0c︡{"stdout": "[x00 x01 x02 x03 x04]\n[x10 x11 x12 x13 x14]\n[x20 x21 x22 x23 x24]\n[x30 x31 x32 x33 x34]\n[x40 x41 x42 x43 x44]"}︡
︠ee606251-ae6a-4b93-b16b-0759aae53fc5︠
%time
c=A.det()
︡ea5a779e-17a1-4873-82ef-56162cdcb264︡{"stdout": "CPU time: 0.00 s,  Wall time: 0.00 s"}︡
︠39cbb88a-0d68-4635-a11f-8cf16e3ddb4c︠
%time
b=det(A)
︡b3a091e5-d22a-472d-8ce5-7f7646487ae9︡{"stdout": "CPU time: 0.05 s,  Wall time: 0.05 s"}︡
︠294b0709-d100-4373-b552-2e252663b746︠
%time
b.expand()
︡e1248b06-6650-497b-ab27-a200cdf484bb︡︡
